#include "win.h"
extern "C" {
#include "coregraphicsservices.h"
}

Win::Win(ProcessSerialNumber psn0, pid_t pid0, string title0, bool focused0,
         int x0, int y0, int w0, int h0)
    : win(0), psn(psn0), pid(pid0), title(title0),
      focused(focused0), visible(true), front(true),
      x(x0), y(y0), w(w0), h(h0) {
    cx = x + w / 2;
    cy = y + h / 2;
}

Win::Win(AXUIElementRef winRef) : win(winRef) {
    CFStringRef titleRef;
    AXUIElementCopyAttributeValue(
        winRef, kAXTitleAttribute, (const void**)&titleRef);
    if (!titleRef) return;
    char buf[1024];
    if (!CFStringGetCString(titleRef, buf, 1023,
                            kCFStringEncodingUTF8)) return;
    title = string(buf);
    if (!title[0]) return;

    CFTypeRef posRef;
    AXUIElementCopyAttributeValue(
        winRef, kAXPositionAttribute, &posRef);
    CGPoint pos;
    AXValueGetValue(
        (const __AXValue*)posRef, kAXValueCGPointType, &pos);
    x = pos.x;
    y = pos.y;

    CFTypeRef sizeRef;
    AXUIElementCopyAttributeValue(
        winRef, kAXSizeAttribute, &sizeRef);
    CGSize size;
    AXValueGetValue(
        (const __AXValue*)sizeRef, kAXValueCGSizeType, &size);
    w = size.width;
    h = size.height;

    CFBooleanRef focusedRef;
    AXUIElementCopyAttributeValue(
        winRef, kAXMainAttribute, (const void**)&focusedRef);
    focused = false;
    if (focusedRef) focused = CFBooleanGetValue(focusedRef);

    cx = x + w / 2;
    cy = y + h / 2;
}

bool Win::hasSameGeom(const Win& r) {
    return x == r.x && y == r.y && w == r.w && h == r.h && title == r.title;
}

void Win::dump() const {
    printf("pid=%d title=%s (%s%s%s):\n"
           "\t(%d,%d) (%d,%d)\n",
           (int)pid, title.c_str(),
           visible?"V":"", front?"F":"", focused?"K":"",
           x, y, w, h);
}

namespace {
    ProcessSerialNumber psn_ = {kNoProcess, kNoProcess};
    CGSConnection cid_ = 0;
    vector<Win> wins_;
    int screenW_ = 0, screenH_ = 0;
    int debug_ = 0;
    vector<pid_t> pids_;

    string getCppString(CFStringRef s) {
        if (!s) return "";
        CFIndex len = CFStringGetLength(s);
        CFIndex max = CFStringGetMaximumSizeForEncoding(
            len, kCFStringEncodingUTF8);
        char* m = (char*)malloc(max + 1);
        CFStringGetCString(s, m, max + 1, kCFStringEncodingUTF8);
        string r(m);
        free(m);
        return r;
    }
}

Mgr::Mgr(int argc, char* argv[]) {
    if (argc > 1 && !strcmp(argv[1], "-d")) debug_ = 1;
}

void Mgr::initWindows(SkipStrategy skipStrategy) {
    //CGSValue titleQuery = CGSCreateCStringNoCopy("kCGSWindowTitle");
    //CGSValue titleQuery = CGSCreateCString("kCGSWindowTitle");
    CFStringRef titleQuery = CFStringCreateWithCStringNoCopy(
        NULL, "kCGSWindowTitle", kCFStringEncodingMacRoman, kCFAllocatorNull);

    CGDirectDisplayID dispId = CGMainDisplayID();
    screenW_ = CGDisplayPixelsWide(dispId);
    screenH_ = CGDisplayPixelsHigh(dispId);

    ProcessSerialNumber frontPSN;
    GetFrontProcess(&frontPSN);
    pid_t pid, frontPID;
    GetProcessPID(&frontPSN, &frontPID);

    ProcessSerialNumber psn = {kNoProcess, kNoProcess};

    GetCurrentProcess(&psn_);
    CGSGetConnectionIDForPSN(0, &psn_, &cid_);

    int focusedID = 0;
    {
        CGSConnection cid = _CGSDefaultConnection();
        int cnt;
        CGSGetOnScreenWindowCount(cid, 0, &cnt);
        int* ids = (int*)malloc(sizeof(int)*cnt);
        CGSGetOnScreenWindowList(cid, 0, cnt, ids, &cnt);
        for (int i = 0; i < cnt; i++) {
            //CGSValue titleVal;
            CFStringRef titleVal;
            CGSGetWindowProperty(cid, ids[i], titleQuery, &titleVal);
            //title = CGSCStringValue(titleVal);
            string title = getCppString(titleVal);
            if (!title.empty() && title != "Shared Menubar") {
                focusedID = ids[i];
                break;
            }
        }
        free(ids);
    }

    for (; GetNextProcess(&psn) == noErr; ) {
        // isVisible? from CGS
        vector<Win> visibles;

        CGSConnection cid;
        if (CGSGetConnectionIDForPSN(0, &psn, &cid) != noErr) continue;

        int cnt;
        if (CGSGetWindowCount(cid_, cid, &cnt) != noErr) continue;
        if (cnt == 0) continue;
        vector<int> ids(cnt);
        CGSGetWindowList(cid_, cid, cnt, &ids[0], &cnt);
        int titleCnt = 0;
        for (int i = 0; i < cnt; i++) {
            int w = ids[i];

            //CGSValue titleVal;
            CFStringRef titleVal;
            CGSGetWindowProperty(cid_, w, titleQuery, &titleVal);
            //char* title = CGSCStringValue(titleVal);
            string title = getCppString(titleVal);
            if (title.empty() || !title[1]) continue;
            titleCnt++;
        }
        if (titleCnt == 0) continue;

        if (CGSGetOnScreenWindowCount(cid_, cid, &cnt) != noErr) continue;
        if (cnt != 0) {
            vector<int> ids(cnt);
            CGSGetOnScreenWindowList(cid_, cid, cnt, &ids[0], &cnt);

            for (int i = 0; i < cnt; i++) {
                int w = ids[i];
                //CGSValue titleVal;
                CFStringRef titleVal;
                CGSGetWindowProperty(cid_, w, titleQuery, &titleVal);
                //char* title = CGSCStringValue(titleVal);
                string title = getCppString(titleVal);
                if (title.empty() || !title[1]) continue;

                CGRect rect;
                CGSGetScreenRectForWindow(cid_, w, &rect);

                pid_t pid;
                CGSConnectionGetPID(cid, &pid, cid_);
                Win win(psn, pid, title, focusedID == w,
                        rect.origin.x, rect.origin.y,
                        rect.size.width, rect.size.height);
                pids_.push_back(pid);
                if (skipStrategy == SKIP_APPLE_EVENT) {
                    wins_.push_back(win);
                }
                else {
                    visibles.push_back(win);
                }
            }
        }

        if (skipStrategy == SKIP_APPLE_EVENT) continue;

        // it makes faster
        if (skipStrategy == SKIP_NON_VISIBLE && visibles.size() == 0) continue;

        // Other infos from Apple Event

        GetProcessPID(&psn, &pid);

        AXUIElementRef appRef = AXUIElementCreateApplication(pid);

        CFArrayRef winRefs;
        AXUIElementCopyAttributeValues(appRef, kAXWindowsAttribute,
                                       0, 255, &winRefs);
        if (!winRefs) continue;

        for (int i = 0; i < CFArrayGetCount(winRefs); i++) {
            AXUIElementRef winRef =
                (AXUIElementRef)CFArrayGetValueAtIndex(winRefs, i);
            Win win(winRef);

            if (!win.title[0]) continue;

            win.psn = psn;
            win.pid = pid;

            win.visible = false;
            for (int j = 0; j < visibles.size(); j++) {
                if (visibles[j].hasSameGeom(win)) {
                    win.visible = true;
                    break;
                }
            }

            win.front = frontPID == pid;

            wins_.push_back(win);
        }
    }
}

void Mgr::searchWindow(Win& w) {
    bool ok = true;
    int cx = w.cx;
    int cy = w.cy;
    for (vector<Win>::const_iterator iter = wins_.begin();
         iter != wins_.end();
         ++iter)
    {
        const Win& win = *iter;
        if (&w == &win) continue;
        if (win.x < cx && win.x + win.w > cx &&
            win.y < cy && win.y + win.h > cy)
        {
            ok = false;
            break;
        }
    }

    if (ok) {
        AXUIElementRef sys = AXUIElementCreateSystemWide();
        AXUIElementRef winRef;
        if (AXUIElementCopyElementAtPosition(sys, cx, cy, &winRef) ==
            kAXErrorSuccess)
        {
            while (winRef) {
                Win c(winRef);
                if (c.title[0]) {
                    w.win = c.win;
                    return;
                }
                AXUIElementCopyAttributeValue(winRef, kAXParentAttribute,
                                              (CFTypeRef*)&winRef);
            }
        }
    }

    for (vector<pid_t>::const_iterator iter = pids_.begin();
         iter != pids_.end();
         ++iter)
    {
        AXUIElementRef appRef = AXUIElementCreateApplication(*iter);

        CFArrayRef winRefs;
        AXUIElementCopyAttributeValues(appRef, kAXWindowsAttribute,
                                       0, 255, &winRefs);
        if (!winRefs) return;

        for (int i = 0; i < CFArrayGetCount(winRefs); i++) {
            AXUIElementRef winRef =
                (AXUIElementRef)CFArrayGetValueAtIndex(winRefs, i);
            Win win(winRef);

            if (!win.title[0]) continue;

            if (w.hasSameGeom(win)) {
                w.win = win.win;
                return;
            }
        }
    }
}

void Mgr::dump() {
    printf("Screen size: (%d, %d)\n", screenW_, screenH_);
    for (int i = 0; i < wins_.size(); i++) {
        wins_[i].dump();
    }
}

vector<Win> Mgr::windows() {
    return wins_;
}

vector<Win> Mgr::visibleWindows() {
    vector<Win> r;
    for (int i = 0; i < wins_.size(); i++) {
        if (wins_[i].visible) r.push_back(wins_[i]);
    }
    return r;
}

int Mgr::focusedWindow() {
    for (int i = 0; i < wins_.size(); i++) {
        Win& w(wins_[i]);
        if (w.visible && w.focused && w.front) return i;
    }
    return -1;
}

void Mgr::switch_to(int index) {
    Win& w(wins_[index]);

    if (debug_) w.dump();

    if (w.win) {
        AXUIElementPerformAction(w.win, kAXRaiseAction);
    }
    else {
        searchWindow(w);
        if (w.win) {
            AXUIElementPerformAction(w.win, kAXRaiseAction);
        }
    }
    SetFrontProcessWithOptions(&w.psn, 1);
}

void Mgr::switch_dir(int x, int y) {
    vector<Win> visibles = visibleWindows();
    int cnt = visibles.size();
    if (cnt == 0) return;
    else if (cnt == 1) {
        // there are no focus window, but we may want to move somewhere
        switch_to(0);
        return;
    }

    int fid = focusedWindow();
    if (fid < 0) {
        switch_to(0);
        return;
    }

    int sx = wins_[fid].cx;
    int sy = wins_[fid].cy;

    int maxDiff = INT_MIN;
    int maxIndex = -1;
    for (int i = 0; i < wins_.size(); i++) {
        if (i == fid) continue;
        if (!wins_[i].visible) continue;

        int dx = wins_[i].cx;
        int dy = wins_[i].cy;
        int diff;
        if (x == 0) {
            int p = (dy - sy) * y;
            if (p > 0)
                diff = 2000 - p - abs(dx - sx) * 3;
            else
                diff = p - abs(dx - sx) * 3;
        }
        else if (y == 0) {
            int p = (dx - sx) * x;
            if (p > 0)
                diff = 2000 - p - abs(dy - sy) * 3;
            else
                diff = p * x - abs(dy - sy) * 3;
        }
        else {
            int px = (dx - sx) * x;
            int py = (dy - sy) * y;
            int ret = 0;
            if (px > 0) ret += 2000 - px;
            else ret += px;
            if (py > 0) ret += 2000 - py;
            else ret += py;
            ret -= abs(px - py) * 3;
            diff = ret;
        }

        if (debug_ ) {
            wins_[i].dump();
            printf("diff=%d\n", diff);
        }

        if (diff > maxDiff) {
            maxDiff = diff;
            maxIndex = i;
        }
    }

    if (debug_) {
        printf("switch to #%d\n", maxIndex);
    }

    switch_to(maxIndex);
}

void Mgr::switch_u() {
    switch_dir(0, -1);
}

void Mgr::switch_r() {
    switch_dir(1, 0);
}

void Mgr::switch_d() {
    switch_dir(0, 1);
}

void Mgr::switch_l() {
    switch_dir(-1, 0);
}

void Mgr::switch_ur() {
    switch_dir(1, -1);
}

void Mgr::switch_dr() {
    switch_dir(1, 1);
}

void Mgr::switch_dl() {
    switch_dir(-1, 1);
}

void Mgr::switch_ul() {
    switch_dir(-1, -1);
}

void Mgr::switch_self() {
    int fid = focusedWindow();
    if (fid >= 0) switch_to(fid);
}

void Mgr::move_to(int wid, int x, int y) {
    Boolean ok;
    AXUIElementIsAttributeSettable(wins_[wid].win, kAXPositionAttribute, &ok);
    if (!ok) return;

    CGPoint pos;
    pos.x = x;
    pos.y = y;
    CFTypeRef posRef = AXValueCreate(kAXValueCGPointType, &pos);

    if (posRef) {
        CGPoint pos;
        AXValueGetValue(
            (const __AXValue*)posRef, kAXValueCGPointType, &pos);

        AXUIElementSetAttributeValue(
            wins_[wid].win, kAXPositionAttribute, posRef);
        CFRelease(posRef);
    }
}

void Mgr::move_rel(int x, int y) {
    int fid = focusedWindow();
    if (fid < 0) return;
    move_to(fid, wins_[fid].x+x, wins_[fid].y+y);
}

void Mgr::move_side(int x, int y) {
    int fid = focusedWindow();
    if (fid < 0) return;
    x *= screenW_ - wins_[fid].w;
    y *= screenH_ - wins_[fid].h;
    if (y == 0) y = MENUBAR_HEIGHT;
    move_to(fid, x, y);
}

void Mgr::move_ur() {
    move_side(1, 0);
}

void Mgr::move_dr() {
    move_side(1, 1);
}

void Mgr::move_dl() {
    move_side(0, 1);
}

void Mgr::move_ul() {
    move_side(0, 0);
}

void Mgr::move_u() {
    move_rel(0, -MOVE_REL_DISTANCE);
}

void Mgr::move_r() {
    move_rel(MOVE_REL_DISTANCE, 0);
}

void Mgr::move_d() {
    move_rel(0, MOVE_REL_DISTANCE);
}

void Mgr::move_l() {
    move_rel(-MOVE_REL_DISTANCE, 0);
}

void Mgr::push_button(CFStringRef s) {
    int fid = focusedWindow();
    if (fid < 0) return;

    AXUIElementRef button;
    AXUIElementCopyAttributeValue(
        wins_[fid].win, s, (const void**)&button);
    if (!button) return;

    AXUIElementPerformAction(button, kAXPressAction);
}

void Mgr::maximize() {
    push_button(kAXZoomButtonAttribute);
}

void Mgr::iconify() {
    push_button(kAXMinimizeButtonAttribute);
}

void Mgr::close() {
    push_button(kAXCloseButtonAttribute);
}

void Mgr::deiconify() {
    for (int i = 0; i < wins_.size(); i++) {
        if (wins_[i].visible) continue;

        Win& w(wins_[i]);
        CFBooleanRef b;
        AXUIElementCopyAttributeValue(
            w.win, kAXMinimizedAttribute, (const void**)&b);
        if (!b) continue;
        if ( CFBooleanGetValue(b)) {
            switch_to(i);
            return;
        }
    }
}
